# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: triple_index.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='triple_index.proto',
  package='',
  syntax='proto3',
  serialized_pb=_b('\n\x12triple_index.proto\"1\n\x0fTranslationUnit\x12\x0f\n\x07\x65lement\x18\x01 \x01(\t\x12\r\n\x05index\x18\x02 \x01(\x04\"V\n\x0bTranslation\x12\"\n\x08\x65ntities\x18\x01 \x03(\x0b\x32\x10.TranslationUnit\x12#\n\trelations\x18\x02 \x03(\x0b\x32\x10.TranslationUnitb\x06proto3')
)




_TRANSLATIONUNIT = _descriptor.Descriptor(
  name='TranslationUnit',
  full_name='TranslationUnit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='element', full_name='TranslationUnit.element', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='index', full_name='TranslationUnit.index', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=71,
)


_TRANSLATION = _descriptor.Descriptor(
  name='Translation',
  full_name='Translation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='entities', full_name='Translation.entities', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='relations', full_name='Translation.relations', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=73,
  serialized_end=159,
)

_TRANSLATION.fields_by_name['entities'].message_type = _TRANSLATIONUNIT
_TRANSLATION.fields_by_name['relations'].message_type = _TRANSLATIONUNIT
DESCRIPTOR.message_types_by_name['TranslationUnit'] = _TRANSLATIONUNIT
DESCRIPTOR.message_types_by_name['Translation'] = _TRANSLATION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TranslationUnit = _reflection.GeneratedProtocolMessageType('TranslationUnit', (_message.Message,), dict(
  DESCRIPTOR = _TRANSLATIONUNIT,
  __module__ = 'triple_index_pb2'
  # @@protoc_insertion_point(class_scope:TranslationUnit)
  ))
_sym_db.RegisterMessage(TranslationUnit)

Translation = _reflection.GeneratedProtocolMessageType('Translation', (_message.Message,), dict(
  DESCRIPTOR = _TRANSLATION,
  __module__ = 'triple_index_pb2'
  # @@protoc_insertion_point(class_scope:Translation)
  ))
_sym_db.RegisterMessage(Translation)


# @@protoc_insertion_point(module_scope)



"""data utils function"""

from collections import defaultdict
from itertools import filterfalse
from functools import reduce
import numpy as np
import logging
import kgedata
from superminhash import Superminhash
from bidict import bidict


class Indexer(object):
    """An indexer for triples."""

    def __init__(self):
        self._ents = bidict()
        self._rels = bidict()
        self._ent_id = 0
        self._rel_id = 0

    def __call__(self, t):
        for e in (t.head, t.tail):
            if e not in self._ents:
                self._ents[e] = self._ent_id
                self._ent_id += 1
        if t.relation not in self._rels:
            self._rels[t.relation] = self._rel_id
            self._rel_id += 1

        return kgedata.TripleIndex(self._ents[t.head], self._rels[t.relation], self._ents[t.tail])

    def mappings(self):
        return self._ents, self._rels

    class UnionFind(object):
        def __init__(self, max_id):
            """every elements has a root of its precedence except the first."""
            self._root = [0] + [i for i in range(max_id-1)]
            self._max_id = max_id

        def add(self, e):
            """Create new root"""
            self._root[e] = e

        def root(self, e):
            return e == self._root[e]

        def roots(self):
            return [self.root(e) for e in range(self._max_id)]

        def union(self, e1, e2):
            """union by the smallest root"""
            root1 = self.find(e1)
            root2 = self.find(e2)
            if root1 == root2: return
            if root1 < root2:
                self._root[root2] = root1
            else:
                self._root[root1] = root2

        def find(self, e):
            """uses path compression."""
            if e <= 0: return 0
            if e != self._root[e]:
                self._root[e] = self.find(self._root[e])
            return self._root[e]

    def shrink_indexes_in_place(self, triples):
        """Uses a union find to find segment."""

        _ent_roots = self.UnionFind(self._ent_id)
        _rel_roots = self.UnionFind(self._rel_id)

        for t in triples:
            _ent_roots.add(t.head)
            _ent_roots.add(t.tail)
            _rel_roots.add(t.relation)

        for i, t in enumerate(triples):
            h = _ent_roots.find(t.head)
            r = _rel_roots.find(t.relation)
            t = _ent_roots.find(t.tail)
            triples[i] = kgedata.TripleIndex(h, r, t)

        ents = bidict()
        available_ent_idx = 0
        for previous_idx, ent_exist in enumerate(_ent_roots.roots()):
            if not ent_exist:
                self._ents.inverse.pop(previous_idx)
            else:
                ents[self._ents.inverse[previous_idx]] = available_ent_idx
            available_ent_idx += 1
        rels = bidict()
        available_rel_idx = 0
        for previous_idx, rel_exist in enumerate(_rel_roots.roots()):
            if not rel_exist:
                self._rels.inverse.pop(previous_idx)
            else:
                rels[self._rels.inverse[previous_idx]] = available_rel_idx
            available_rel_idx += 1
        self._ents = ents
        self._rels = rels
        self._ent_id = available_ent_idx
        self._rel_id = available_rel_idx


def build_index_and_mapping(triples):
    """index all triples into indexes and return their mappings"""
    ents = bidict()
    rels = bidict()
    ent_id = 0
    rel_id = 0

    collected = []
    for t in triples:
        for e in (t.head, t.tail):
            if e not in ents:
                ents[e] = ent_id
                ent_id += 1
        if t.relation not in rels:
            rels[t.relation] = rel_id
            rel_id += 1
        collected.append(kgedata.TripleIndex(ents[t.head], rels[t.relation], ents[t.tail]))

    return collected, ents, rels


def recover_triples_from_mapping(indexes, ents: bidict, rels: bidict):
    """recover triples from mapping."""
    triples = []
    for t in indexes:
        triples.append(kgedata.Triple(ents.inverse[t.head], rels.inverse[t.relation], ents.inverse[t.tail]))
    return triples


def unpack(triple):
    """unpacks triple into (h, r, t). Can take Triple or TripleIndex."""
    return triple.head, triple.relation, triple.tail

def shuffle_triples(triples):
    """Shuffle triples."""
    return np.random.permutation(triples)

def split_golden_set(triples, valid_ratio, test_ratio):
    """Split the data into train/valid/test sets."""
    assert valid_ratio >= 0.0
    assert test_ratio >= 0.0
    num_valid = int(len(triples) * valid_ratio)
    num_test = int(len(triples) * test_ratio)
    valid_set = triples[:num_valid]
    test_set = triples[num_valid:num_valid+num_test]
    train_set = triples[num_valid+num_test:]
    assert len(valid_set) + len(test_set) + len(train_set) == len(triples)

    return train_set, valid_set, test_set

def _transform_triple_numpy(x):
    """Transform triple index into a 1-D numpy array."""
    return np.array([x.head, x.relation, x.tail], dtype=np.int64)

def _generate_triple_from_list(triples):
    for t in triples:
        yield t.head
        yield t.relation
        yield t.tail

def pack_triples_numpy(triples):
    """Packs a list of triple indexes into a 2D numpy array."""
    return np.fromiter(_generate_triple_from_list(triples), dtype=np.dtype(np.int64), count=len(triples)).reshape(-1, 3)

def _set_close_to(a, b, threshold):
    return a.similarity(b) > threshold

def _assert_threshold(threshold):
    if threshold <= 0.0 or threshold > 1.0: raise ValueError("The threshold {} is not in the valid range of (0, 1]")

def remove_deficit_relation(triples, threshold=1000):
    logging.debug("remove deficit relation")

    counter = defaultdict(int)
    for t in triples:
        counter[t.relation] += 1

    num_triples = len(triples)
    removal_set = set()
    for rel, rel_items in counter.items():
        if rel_items < threshold:
            removal_set.add(rel)
    logging.info("Removing {} relations: {}".format(len(removal_set), str(removal_set)))

    return list(filterfalse(lambda x: x.relation in removal_set, triples))

def remove_near_duplicate_relation(triples, threshold=0.97):
    """If entity pairs in a relation is as close as another relations, only keep one relation of such set."""
    logging.debug("remove duplicate")

    _assert_threshold(threshold)

    duplicate_rel_counter = defaultdict(list)
    relations = set()
    for t in triples:
        duplicate_rel_counter[t.relation].append(f"{t.head} {t.tail}")
        relations.add(t.relation)
    relations = list(relations)

    num_triples = len(triples)
    removal_relation_set = set()

    for rel, values in duplicate_rel_counter.items():
        duplicate_rel_counter[rel] = Superminhash(values)
    for i in relations:
        for j in relations:
            if i == j or i in removal_relation_set or j in removal_relation_set: continue
            close_relations = [i]
            if _set_close_to(duplicate_rel_counter[i], duplicate_rel_counter[j], threshold):
                close_relations.append(j)
        if len(close_relations) > 1:
            close_relations.pop(np.random.randint(len(close_relations)))
            removal_relation_set |= set(close_relations)
    logging.info("Removing {} relations: {}".format(len(removal_relation_set), str(removal_relation_set)))

    return list(filterfalse(lambda x: x.relation in removal_relation_set, triples))

def remove_inverse_relation(triples, threshold=0.97):
    """if entity pairs in a relation overlaps its reverse pairs in another relation, only keep one relation of such set."""
    logging.debug("remove inverse")
    _assert_threshold(threshold)

    rel_counter = defaultdict(list)
    inverse_rel_counter = defaultdict(list)
    relations = set()
    for t in triples:
        rel_counter[t.relation].append(f"{t.head} {t.tail}")
        inverse_rel_counter[t.relation].append(f"{t.tail} {t.head}")
        relations.add(t.relation)
    relations = list(relations)

    for rel, values in rel_counter.items():
        rel_counter[rel] = Superminhash(values)
    for rel, values in inverse_rel_counter.items():
        inverse_rel_counter[rel] = Superminhash(values)

    num_triples = len(triples)
    removal_relation_set = set()
    for i in relations:
        if i in removal_relation_set: continue
        close_relations = [i]
        for j in relations:
            if i == j or j in removal_relation_set: continue
            if _set_close_to(rel_counter[i], inverse_rel_counter[j], threshold):
                close_relations.append(j)
        if len(close_relations) > 1:
            close_relations.pop(np.random.randint(len(close_relations)))
            removal_relation_set |= set(close_relations)
    logging.info("Removing {} relations: {}".format(len(removal_relation_set), str(removal_relation_set)))

    return list(filterfalse(lambda x: x.relation in removal_relation_set, triples))

def remove_direct_link_triples(train, valid, test):
    """Remove direct links in the training sets."""
    pairs = set()
    merged = valid + test
    for t in merged:
        pairs.add((t.head, t.tail))

    filtered = filterfalse(lambda t: (t.head, t.tail) in pairs or (t.tail, t.head) in pairs, train)
    return list(filtered)

def remove_unqualified_relations_from_triples(triples, deficit_threshold, duplicate_threshold, inverse_threshold):
    triples = remove_deficit_relation(triples, deficit_threshold)
    triples = remove_near_duplicate_relation(triples, duplicate_threshold)
    triples = remove_inverse_relation(triples, inverse_threshold)
    logging.info("current triples: " + str(len(triples)))
    return triples

def build_dataset(triples, valid_ratio, test_ratio):
    triples = list(shuffle_triples(triples))
    train, valid, test = split_golden_set(triples, valid_ratio, test_ratio)
    train = remove_direct_link_triples(train, valid, test)
    return train, valid, test



"""Shuffle and split the data into train/valid/test sets."""

import sys
import logging
import kgekit.data
import kgekit.io
import kgedata
from bidict import bidict

logging.basicConfig(level=logging.DEBUG)


def build_dataset(filename, deficit_threshold=1000, delimiter='\t', triple_order="hrt"):
    indexer = kgekit.data.Indexer()
    triples = []
    for triple in kgekit.io.yield_triples(filename, triple_order=triple_order, delimiter=delimiter):
        triples.append(indexer(triple))

    logging.info("removing unqualified relations")
    triples = kgekit.data.remove_unqualified_relations_from_triples(triples, deficit_threshold, duplicate_threshold=0.97, inverse_threshold=0.97)

    logging.info("update indexes")
    indexer.shrink_indexes_in_place(triples)
    ents, rels = indexer.mappings()

    logging.info("recording translations")
    kgekit.io.write_index_translation("translation.protobuf", ents, rels)

    logging.info("build dataset")
    train, valid, test = kgekit.data.build_dataset(
        triples,
        0.005,
        0.005)
    logging.info("writing triple files")

    kgekit.io.write_triples('train.txt', train, delimiter=delimiter, triple_order=triple_order)
    kgekit.io.write_triples('valid.txt', valid, delimiter=delimiter, triple_order=triple_order)
    kgekit.io.write_triples('test.txt', test, delimiter=delimiter, triple_order=triple_order)

if __name__ == '__main__':
    if len(sys.argv) > 3:
        build_dataset(sys.argv[1], float(sys.argv[2]))
    else:
        build_dataset(sys.argv[1])



"""IO functions"""

from pathlib import Path
import kgekit.data
import kgedata
import kgekit.triple_index_pb2 as triple_pb
from kgekit.utils import assert_triple_order

DEFAULT_DELIMITER=' '

def _assert_good_file(filename):
    try:
        path = Path(filename).resolve(strict=True)
    except FileNotFoundError:
        raise FileNotFoundError("Can't find triple file at path {}".format(filename))
    except RuntimeError:
        raise FileNotFoundError("Cyclic dependency when find triple file {}".format(filename))

    if not (path.exists() and path.is_file()):
        raise FileNotFoundError("Can't find the triple file at {}.".format(path))

def yield_triples(filename, triple_order="hrt", delimiter=DEFAULT_DELIMITER, read_fn=kgedata.get_triple):
    _assert_good_file(filename)
    assert_triple_order(triple_order)
    with open(filename) as f:
        line = f.readline().rstrip('\n')
        while line:
            triple = read_fn(line, triple_order, delimiter, True)
            if triple is not None:
                yield triple
            line = f.readline().rstrip('\n')

def read_triples(filename, read_fn=kgedata.get_triple, triple_order="hrt", delimiter=DEFAULT_DELIMITER, skip_first_line=False):
    _assert_good_file(filename)
    assert_triple_order(triple_order)
    with open(filename) as f:
        if skip_first_line:
            next(f)
        indexes = [read_fn(l.rstrip('\n'), triple_order, delimiter, True) for l in f]
        filtered = list(filter(None.__ne__, indexes))
        num_none = len(indexes) - len(filtered)
        return filtered, num_none

def read_triple_indexes(filename, read_fn=kgedata.get_triple_index, triple_order="hrt", delimiter=DEFAULT_DELIMITER, skip_first_line=False):
    return read_triples(filename, read_fn, triple_order, delimiter, skip_first_line)

def read_triple_indexes_numpy(filename, triple_order="hrt", delimiter=DEFAULT_DELIMITER, skip_first_line=False):
    indexes, num_failed = read_triple_indexes(
        filename,
        triple_order=triple_order,
        delimiter=delimiter,
        skip_first_line=skip_first_line)
    return kgekit.data.pack_triples_numpy(indexes), num_failed

def _label_processing(l, delimiter):
    splits = l.rstrip().split(delimiter)
    return (splits[0], delimiter.join(splits[1:]))

# TODO: This consumes around 26G memory. It will be nicer to have stream version.
def read_labels(filename, delimiter=DEFAULT_DELIMITER):
    """read label files. Format: ent label"""
    _assert_good_file(filename)
    with open(filename) as f:
        labels = [_label_processing(l, delimiter) for l in f]
        return labels

def _label_processing(l, delimiter):
    splits = l.rstrip().split(delimiter)
    return (splits[0], delimiter.join(splits[1:]))

def _desc_processing(l, delimiter):
    splits = l.rstrip().split(delimiter)
    desc = delimiter.join(splits[1:])
    desc_splits = desc.split('@')
    return (splits[0], '@'.join(desc_splits[:-2]), desc_splits[-1])

# see `read_labels`
def read_descriptions(filename, delimiter=DEFAULT_DELIMITER):
    """read label files. Format: ent label"""
    _assert_good_file(filename)
    with open(filename) as f:
        descs = [_desc_processing(l, delimiter) for l in f]
        return descs


def write_index_translation(translation_filename, entity_ids, relation_ids):
    """write triples into a translation file."""
    translation = triple_pb.Translation()
    entities = []
    for name, index in entity_ids.items():
        translation.entities.add(element=name, index=index)
    relations = []
    for name, index in relation_ids.items():
        translation.relations.add(element=name, index=index)
    with open(translation_filename, "wb") as f:
        f.write(translation.SerializeToString())

def write_triples(filename, triples, delimiter=DEFAULT_DELIMITER, triple_order="hrt"):
    """write triples to file."""
    with open(filename, 'w') as f:
        for t in triples:
            line = t.serialize(delimiter, triple_order)
            f.write(line + "\n")

def read_translation(filename):
    """Returns protobuf mapcontainer. Read from translation file."""
    translation = triple_pb.Translation()
    with open(filename, "rb") as f:
        translation.ParseFromString(f.read())

    def unwrap_translation_units(units):
        for u in units: yield u.element, u.index

    return (list(unwrap_translation_units(translation.entities)),
        list(unwrap_translation_units(translation.relations)))

def read_openke_translation(filename, delimiter='\t', entity_first=True):
    """Returns map with entity or relations from plain text."""
    result = {}
    with open(filename, "r") as f:
        _ = next(f) # pass the total entry number
        for line in f:
            line_slice = line.rstrip().split(delimiter)
            if not entity_first:
                line_slice = list(reversed(line_slice))
            result[line_slice[0]] = line_slice[1]

    return result



import sys
from collections import defaultdict
import kgekit.io
import kgekit.data
import random
from itertools import chain, dropwhile

class Transform(object):
    def __init__(self, entities):
        self.entities = entities

    def __call__(self, label):
        if len(label) != 2:
            print(label)
            return True
        name, _ = label
        return name not in self.entities

def match_triple_labels(label_filename, translation_filename):
    entities, _ = kgekit.io.read_translation(translation_filename)
    labels = kgekit.io.read_labels(label_filename, ' ')
    t = Transform(entities)
    filtered_labels = dropwhile(t, labels)
    filtered_labels = list(map(lambda l: " ".join((str(entities[l[0]]), l[1])), filtered_labels))
    return filtered_labels

if __name__ == '__main__':
    labels = match_triple_labels(sys.argv[1], sys.argv[2])
    with open('matched_labels.txt', 'w') as f:
        for l in labels:
            f.write(l + "\n")



import sys
from collections import defaultdict
import kgekit.io
import kgekit.data
import random
from itertools import chain, dropwhile

class Transform(object):
    def __init__(self, entities, lang):
        self.entities = entities
        self.lang = lang

    def __call__(self, description):
        if len(description) != 3:
            print(description)
            return True
        name, _, lang = description
        return lang != self.lang or name not in self.entities

def match_triple_descriptions(description_filename, translation_filename, lang):
    entities, _ = kgekit.io.read_translation(translation_filename)
    descriptions = kgekit.io.read_descriptions(description_filename, '\t')
    t = Transform(entities)
    filtered_descriptions = dropwhile(t, descriptions)
    filtered_descriptions = list(map(lambda l: " ".join((str(entities[l[0]]), l[1])), filtered_descriptions))
    return filtered_descriptions

if __name__ == '__main__':
    descriptions = match_triple_descriptions(sys.argv[1], sys.argv[2], sys.argv[3])
    with open('matched_descriptions.txt', 'w') as f:
        for l in descriptions:
            f.write(l + "\n")



"""YAGO3's ttle file contain escaped unicode symbol like `\u0022`. They are
escaped by `escape_file.py`. This file reverse the symbol of percentage encoded
symbols to UTF-8 string.
"""
import re
import sys
import os
import binascii

def unescape_pecentage_encoded(matchobj):
    if matchobj:
        return binascii.unhexlify(matchobj.group(1)).decode('charmap')
    else:
        return ''

def unescape_file(path):
    paths, filename = os.path.split(path)
    with open(path, 'r') as f1, open(os.path.join(paths, "processed_"+filename), 'w') as f2:
        content = f1.read()
        escaped = re.sub(r'%(\d\d)', unescape_pecentage_encoded, content)
        f2.write(escaped)

unescape_file(sys.argv[1])
